import pytest

from . import storage


@pytest.hookspec(firstresult=True)
def pytest_qase_file_storages() -> dict[str, storage.FileStorage]:  # type: ignore
    """Return mapping options to file storage instance.

    Example:
        {
            "qase": QaseFileStorage(),
        }

    """



class BaseQasePluginException(Exception):
    """Represent BaseQasePlugin exception."""

    message: str = ""

    def __init__(self, message: str = "", *args: object) -> None:
        super().__init__(*args)
        if message:
            self.message = message


class InvalidCaseId(BaseQasePluginException):
    """Exception that signifies that incorrect case was set for test."""

    message = "Tests have incorrect cases ids. Please check logs"


class DuplicatingCaseId(BaseQasePluginException):
    """Exception that signifies that incorrect case was set for test."""

    def __init__(self, duplicating_ids: list[int], *args: object) -> None:
        ids = ", ".join(str(i) for i in duplicating_ids)
        super().__init__(
            message=f"Duplicating qase IDs found: {ids}",
            *args,   # type: ignore
        )


class MultipleIDsForTest(BaseQasePluginException):
    """Exception that signifies that single test marked with multiple IDs.

    Each test should be associated with exactly 1 case from qase.io for
    following reasons:

    * Atomic checks. It's easier to debug and fix
    * Adding support of multiple IDs to plugin will make it more complex.

    """

    message = "Multiple qase IDs associated with single test"


class RunNotConfigured(BaseQasePluginException):
    """Exception that signifies that test run not configured."""

    message = "Test run not configured"



import logging
import sys

import qaseio
import qaseio.apis
from qaseio import models


class QaseClient:
    """Class for interacting with tests runs in Qase API."""

    def __init__(
        self,
        token: str,
        project_code: str,
    ):
        """Init client."""
        super().__init__()
        self._logger = logging.getLogger("qase")
        self._logger.addHandler(
            logging.StreamHandler(sys.stderr),
        )
        self._client = qaseio.ApiClient(
            configuration=qaseio.Configuration(
                api_key={
                    "TokenAuth": token,
                },
            ),
        )
        self._project_code: str = project_code

    def get_run(
        self,
        run_id: int,
    ) -> models.Run:
        return qaseio.apis.RunsApi(self._client).get_run(
            code=self._project_code, id=run_id,
        ).result

    def create_run(
        self,
        run_data: models.RunCreate,
    ) -> models.Run:
        """Create test run in Qase."""
        response = qaseio.apis.RunsApi(self._client).create_run(
            code=self._project_code,
            run_create=run_data,
        )
        return self.get_run(response.result.id)

    def load_cases_ids(
        self,
    ) -> list[int]:
        """Load all cases ids of project."""
        limit = 100
        cases: list[int] = []
        response: models.TestCaseListResponse = qaseio.apis.CasesApi(self._client).get_cases(
            code=self._project_code,
            limit=limit,
            offset=len(cases),
        )
        while True:
            response = qaseio.apis.CasesApi(self._client).get_cases(
                code=self._project_code,
                limit=limit,
                offset=len(cases),
            )
            new_cases = [case.id for case in response.result.entities]
            cases += new_cases
            if not new_cases:
                break
        return cases

    def report_test_results(
        self,
        run: models.RunCreate,
        report_data: models.ResultCreate,
    ) -> tuple[str, models.ResultCreate]:
        """Report test results back to Qase."""
        result = qaseio.apis.ResultsApi(self._client).create_result(
            code=self._project_code,
            id=run.id,
            result_create=report_data,
        ).result
        return result.hash, report_data.status



import io
from typing import Protocol

import qaseio
import qaseio.apis


class FileStorage(Protocol):
    """Base class for file uploader."""

    def save_file_obj(self, content: bytes, filename: str) -> str:
        """Upload file to storage and return URL."""


class QaseFileStorage:
    """Upload files to Qase S3 bucket as attachment."""

    def __init__(
        self,
        qase_token: str,
        qase_project_code: str,
    ):
        """Prepare ApiClient for qase io using credentials."""
        self._client = qaseio.ApiClient(
            configuration=qaseio.Configuration(
                api_key={
                    "TokenAuth": qase_token,
                },
            ),
        )
        self._project_code = qase_project_code

    def save_file_obj(self, content: bytes, filename: str) -> str:
        """Upload file to Qase.io S3 bucket via attachment API."""
        file_obj = io.BytesIO(content)
        file_obj.name = filename

        result = qaseio.apis.AttachmentsApi(self._client).upload_attachment(
            code=self._project_code,
            file=[file_obj],
        ).result[0]
        return result["url"]



RUN_NAME_TEMPLATE = "({env}) Automated Test Run {browser} {date}"
REPORT_FOLDER_TEMPLATE = "{env}/{browser}/run-{id}/{test_name}"

TEST_PASSED = "Test Passed"
TEST_FAILED = "Test Failed, on `{when}`"

FAILED_TEST_REPORT_TEMPLATE = """
---

* URL: [URL]({url})
* Browser log: [Browser log]({browser_log_url})
* Screenshot: ![driver screenshot]({screenshot_url})
* HTML: [HTML file]({html_url})

---
"""



import collections
import datetime
import logging
import sys

import pytest
from qaseio import models

from . import constants, debug_info, plugin_exceptions, storage


class QaseConverter:
    """Class for converting data from pytest to qase."""

    def __init__(
        self,
        browser: str,
        env: str,
        project_code: str,
        file_storage: storage.FileStorage | None,
    ):
        """Init converter."""
        super().__init__()
        self._logger = logging.getLogger("qase")
        self._logger.addHandler(
            logging.StreamHandler(sys.stderr),
        )
        self._env = env
        self._browser = browser
        self._project_code = project_code
        self._file_storage = file_storage

    def prepare_run_data(
        self,
        cases_ids_from_api: list[int],
        items: list[pytest.Function],
    ) -> tuple[models.RunCreate, dict[str, int | None]]:
        """Prepare data needed to create test run."""
        cases, tests = self._prepare_cases_for_run(
            cases_ids_from_api=cases_ids_from_api,
            items=items,
        )
        title = constants.RUN_NAME_TEMPLATE.format(
            env=self._env.capitalize(),
            browser=self._browser.capitalize(),
            date=datetime.datetime.utcnow().strftime("%m/%d/%Y %H:%M:%S"),
        )
        run_data = models.RunCreate(
            title=title,
            cases=cases,
        )

        return run_data, tests

    def prepare_report_data(
        self,
        case_id: int,
        run_id: int,
        item: pytest.Function,
        report: pytest.TestReport,
    ) -> models.ResultCreate:
        """Create a test result based on results from pytest."""
        if hasattr(report, "wasxfail"):
            return self._prepare_skipped_test_report(
                case_id=case_id,
                report=report,
            )
        match report.outcome:
            case "passed":
                return self._prepare_passed_test_report(
                    case_id=case_id,
                    report=report,
                )
            case "skipped":
                return self._prepare_skipped_test_report(
                    case_id=case_id,
                    report=report,
                )
            case "failed":
                return self._prepare_failed_test_report(
                    case_id=case_id,
                    run_id=run_id,
                    item=item,
                    report=report,
                )
        raise ValueError("Failed to convert test result!")

    def _prepare_passed_test_report(
        self,
        case_id: int,
        report: pytest.TestReport,
    ) -> models.ResultCreate:
        """Prepare result report for passed test."""
        return models.ResultCreate(
            case_id=case_id,
            status="passed",
            comment=constants.TEST_PASSED,
            time_ms=int(report.duration * 1000),
        )

    def _prepare_skipped_test_report(
        self,
        case_id: int,
        report: pytest.TestReport,
    ) -> models.ResultCreate:
        """Prepare result report for skipped test."""
        if hasattr(report, "wasxfail"):
            skip_reason = report.wasxfail
        else:
            *_, skip_reason = report.longrepr  # type: ignore
        return models.ResultCreate(
            case_id=case_id,
            status="skipped",
            comment=skip_reason,
            time_ms=int(report.duration * 1000),
        )

    def _prepare_failed_test_report(
        self,
        case_id: int,
        run_id: int,
        item: pytest.Function,
        report: pytest.TestReport,
    ) -> models.ResultCreate:
        """Prepare result report for failed test."""
        comment = constants.TEST_FAILED.format(when=report.when)
        debug_information = (
            debug_info.DebugInfo(
                item=item,
                webdriver=item._webdriver,
            )
            if hasattr(item, "_webdriver")
            else None
        )
        if debug_information and self._file_storage:
            folder = constants.REPORT_FOLDER_TEMPLATE.format(
                env=self._env,
                id=run_id,
                browser=self._browser,
                test_name=item.name,
            )
            debug_comment = debug_information.generate_debug_comment(
                file_storage=self._file_storage,
                folder=folder,
            )
            comment += f"\n{debug_comment}"

        return models.ResultCreate(
            case_id=case_id,
            status="failed",
            comment=comment,
            time_ms=int(report.duration * 1000),
            stacktrace=report.longreprtext,
        )

    def _prepare_cases_for_run(
        self,
        cases_ids_from_api: list[int],
        items: list[pytest.Function],
    ) -> tuple[list[int], dict[str, int | None]]:
        """Collect test cases from test markers.

        Raise InvalidCaseId in case if incorrect case id was provided. Raise DuplicatingCaseId in
        case if duplicated cased ids were provided.

        """
        cases_ids = []

        case_id_invalid = False
        # list of parsed markers to track duplicating case IDs in different tests
        parsed_markers_ids = []
        # Mapping of pytest items ids and case id
        tests: dict[str, int | None] = {}
        for item in items:
            case_id = self._extract_case_id_from_test(
                project_code=self._project_code,
                item=item,
            )
            tests[item.nodeid] = case_id
            qase_marker = self._extract_qase_marker(item)
            qase_marker_id = id(qase_marker)
            if qase_marker_id in parsed_markers_ids:
                continue
            parsed_markers_ids.append(qase_marker_id)

            if not case_id:
                self._logger.error(
                    f"No case id not found for {item.name}! "
                    f"Please add case id for {item.location[0]}:1",
                )
                case_id_invalid = True
                continue
            if case_id not in cases_ids_from_api:
                self._logger.error(
                    f"Case with ID {case_id} not found for {item.name}! "
                    f"Please check case id for {item.location[0]}",
                )
                case_id_invalid = True
                continue
            cases_ids.append(case_id)

        duplicating_ids = [
            case_id for case_id, counter in collections.Counter(cases_ids).items()
            if counter > 1
        ]
        if duplicating_ids:
            raise plugin_exceptions.DuplicatingCaseId(duplicating_ids)

        if case_id_invalid:
            raise plugin_exceptions.InvalidCaseId()
        return cases_ids, tests

    def _extract_qase_marker(
        self,
        item: pytest.Function,
    ) -> pytest.Mark | None:
        """Extract qase's `Mark` object from test item."""
        qase_markers = [
            marker for marker in item.own_markers
            if marker.name == "qase"
        ]
        if not qase_markers:
            return None

        if len(qase_markers) == 1:
            return qase_markers[0]

        self._logger.error(
            f"Multiple qase IDs associated with: {item}",
        )
        raise plugin_exceptions.MultipleIDsForTest()

    def _extract_case_id_from_marker(
        self,
        project_code: str,
        marker: pytest.Mark,
    ):
        """Shortcut to extract qase case ID from marker."""
        if len(marker.args) != 1:
            return None
        url = marker.args[0]
        return int(url.split(f"{project_code}-")[-1])

    def _extract_case_id_from_test(
        self,
        project_code: str,
        item: pytest.Function,
    ) -> int | None:
        """Get test case's id marker from test item."""
        marker = self._extract_qase_marker(item=item)
        if not marker:
            return None
        return self._extract_case_id_from_marker(project_code=project_code, marker=marker)
